/*
    chpasswd.c: Change password by /usr/bin/passwd
    Copyleft by SATOH Fumiyasu <fumiya@cij.co.jp>
    since 1997/05/23

    sample: poppasswd.c 1.4
    note:   original comments moved to last of this source
*/

#define BUFSIZE 512

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <ctype.h>
#include <unistd.h>
#include <string.h>
#include <termios.h>
#include <dirent.h>
#include <fcntl.h>
#include <pwd.h>

#include "trace.h"

#include "chpasswd.h"

static int FindPTY(char **p_slave_dev);
static void WriteString(int fd, char *s);
static int Match(char *str,char *pat);
static int Expect(int master_fd,char **expected,char *buf);

static RESULT DoChild(int master_fd,char *slave_dev,char *user);
static RESULT Talk2Child(int master_fd,char *user,char *pass,char *newpass,char *emess);
static void GetErrorMessage(int master_fd,char **expected,char *buf);


/* Prompt strings expected from the "passwd" command. If you want
 * to port this program to yet another flavor of UNIX, you may need to add
 * more prompt strings here.
 *
 * Each prompt is defined as an array of pointers to alternate 
 * strings, terminated by an empty string. In the strings, '*'
 * matches any sequence of 0 or more characters. Pattern matching
 * is case-insensitive.
 */

/* Solaris 2.5.1	passwd, NIS, NIS+ */
/* SunOS 4.1.3-JL	NIS */
/* BSD/OS 2.0.1		passwd */
/* Linux 2.0.31		passwd */

static char *P1[] = {
    /* Solaris 2.x */
    "Enter login password: ",
    "Enter login(NIS) password: ",
    "Enter login(NIS+) password: ",
    /* SunOS 4.x */
    "Changing password for * on *.\nOld password:",
    "Changing NIS password for * on *.\nOld password:",
    /* BSD/OS 2.x */
    "Changing local password for *.\nOld password:",
    /* Linux */
    "Changing password for *\nEnter old password: ",
    ""
};

static char *P2[] = {
    /* Solaris 2.x, SunOS 4.x, BSD/OS 2.x */
    "\nNew password:",
    /* BSD/OS 2.1 */
    "\nNew password (8 significant characters):",
    /* Linux */
    "\nEnter new password: ",
    "\r\nEnter new password: ",
    ""
};


static char *P3[] = {
    /* Solaris 2.x */
    "\nRe-enter new password:",
    /* SunOS 4.x, BSD/OS 2.x */
    "\nRetype new password:",
    /* Linux */
    "\nRe-type new password: ",
    "\r\nRe-type new password: ",
    ""
};
    
static char *P4[] = {
    /* Solaris 2.x */
    "NIS(YP) passwd/attributes changed on *",
    "\n        NIS+ password information changed for *\n",
    /* SunOS 4.x */
    "NIS entry changed on *\n",
    /* BSD/OS 2.x */
    "\npasswd: updating passwd database\n",
    /* Linux */
    "\nPassword changed.\n",
    "\r\nPassword changed.\n",
    /* generic */
    "\n",
    ""
};

RESULT ChangePasswd(user_in, passwd_in, passwd_new)
char *user_in, *passwd_in, *passwd_new;
{
    char	emess[BUFSIZE];
    char	*slave_dev;
    int		master_fd;
    pid_t	pid, wpid;
    int		wstat;

    if (IsSuitablePasswd(passwd_new, user_in, passwd_in) != TRUE) {
	TRACE(1, ("new passwd is not suitable.\n"));
	return FAILURE;
    }

    /* get pty to talk to password program */
    if ((master_fd = FindPTY(&slave_dev)) < 0) {
	TRACE(1, ("unable to find pty.\n"));
	return FAILURE;
    }

    /* fork child process to talk to password program */
    if ((pid = fork()) < 0) {   /* Error, can't fork */
	TRACE_STRERROR(1, ("unable to fork.\n"));
	return FAILURE;
    }

    if (pid) { /* Parent */
	sleep(1);    /* Make sure child is ready.  Is this really needed? */
	if (Talk2Child(master_fd, user_in, passwd_in, passwd_new, emess) == FAILURE) {
	    TRACE(1, ("failed attempt by %s.\n", user_in));
	    if (*emess == '\0') {
	          TRACE(1, ("unable to change password.\n" ));
            } else {
		  TRACE(1, ("%s\n", emess));
            }
	    return FAILURE;
	}

	TRACE(3, ("waiting for child %s. (%d)\n", OPT_PATH_PASSWD, pid));
	if ((wpid = waitpid(pid, &wstat, 0)) < 0) {
	    TRACE_STRERROR(1, ("waiting for child %s failed.\n", OPT_PATH_PASSWD));
	    return FAILURE;
	}

	if (pid != wpid) {
	    TRACE(1, ("wrong child %s waiting for!\n", OPT_PATH_PASSWD));
	    return FAILURE;
	}

	if (WIFEXITED(wstat) == 0) {
	    TRACE(1, ("child %s killed? (%d)\n", OPT_PATH_PASSWD, wstat));
	    return FAILURE;
	}

	if (WEXITSTATUS(wstat) != 0) {
	    TRACE(1, ("child %s exited abnormally. (%d)\n", OPT_PATH_PASSWD, wstat));
	    return FAILURE;
	}
	TRACE(3, ("child %s returned %d.\n", OPT_PATH_PASSWD, WEXITSTATUS(wstat)));

	TRACE(1, ("password changed for %s.\n", user_in));

	return SUCCESS;
    } else {    /* Child */
	/*
	 * Become the user trying who's password is being changed.  We're
	 * about to exec /usr/bin/passwd with is setuid root anyway, but this
	 * way it looks to the child completely like it's being run by
	 * the normal user, which makes it do its own password verification
	 * before doing any thing.  In theory, we've already verified the
	 * password, but this extra level of checking doesn't hurt.  Besides,
	 * the way I do it here, if somebody manages to change somebody
	 * else's password, you can complain to your vendor about security
	 * holes, not to me!
	 */

	struct passwd *normal_pwd;

	normal_pwd = getpwnam(user_in);
	if (normal_pwd == NULL) {
	    TRACE(1, ("unable to get password entry for %s.\n", user_in));
	    return FAILURE;
	}
	setuid(normal_pwd->pw_uid);
	setgid(normal_pwd->pw_gid);
	DoChild(master_fd, slave_dev, user_in);
	exit(0);
    }
}

/*
 * DoChild
 *
 * Do child stuff - set up slave pty and execl /usr/bin/passwd.
 *
 * Code adapted from "Advanced Programming in the UNIX Environment"
 * by W. Richard Stevens.
 *
 */
RESULT DoChild(master_fd, slave_dev, user)
int master_fd;
char *slave_dev, *user;
{
    int slave_fd;
    struct termios stermios;

    /* Start new session - gets rid of controlling terminal. */
    if (setsid() < 0) {
	TRACE_STRERROR(1, ("setsid failed.\n"));
	return FAILURE;
    }

    /* Open slave pty and acquire as new controlling terminal. */
    if ((slave_fd = open(slave_dev, O_RDWR)) < 0) {
	TRACE_STRERROR(1, ("unable to open slave pty.\n"));
	return FAILURE;
    }

#ifdef OSF1
/*
 * under OSF/1, there is no implicit assignment of control terminal with
 * open(). The call to ioctl below opens the slave device and attaches
 * it to the current session as the controlling terminal. Refer to
 * tty(7) under OSF/1 for details.
 */
    ioctl(slave_fd,TIOCSCTTY,0);
#endif

    /* Close master_fd. */
    close(master_fd);

    /* Make slave stdin/out/err of child. */
    if (dup2(slave_fd, STDIN_FILENO) != STDIN_FILENO) {
	TRACE_STRERROR(1, ("dup2 error. (stdin)\n"));
	return FAILURE;
    }
    if (dup2(slave_fd, STDOUT_FILENO) != STDOUT_FILENO) {
	TRACE_STRERROR(1, ("dup2 error. (stdout)\n"));
	return FAILURE;
    }
    if (dup2(slave_fd, STDERR_FILENO) != STDERR_FILENO) {
	TRACE_STRERROR(1, ("dup2 error. (stderr)\n"));
	return FAILURE;
    }
    if (slave_fd > 2) close(slave_fd);

    /* Set proper terminal attributes - no echo, canonical input processing,
	no map NL to CR/NL on output. */
    if (tcgetattr(0, &stermios) < 0) {
	TRACE_STRERROR(1, ("tcgetattr error.\n"));
	return FAILURE;
    }
    stermios.c_lflag &= ~(ECHO | ECHOE | ECHOK | ECHONL);
    stermios.c_lflag |= ICANON;
    stermios.c_oflag &= ~(ONLCR);
    if (tcsetattr(0, TCSANOW, &stermios) < 0) {
	TRACE_STRERROR(1, ("tcsetattr error.\n"));
	return FAILURE;
    }

    /* Exec /usr/bin/passwd. */
#if OPT_PASSWD_USERNAME == 1
    if (execl(OPT_PATH_PASSWD, "passwd", user, (char*)0) < 0)
#else /* OPT_PASSWD_USERNAME == 1*/
    if (execl(OPT_PATH_PASSWD, "passwd", (char*)0) < 0)
#endif /* OPT_PASSWD_USERNAME == 1*/
    {
	TRACE_STRERROR(1, ("unable to exec %s.\n", OPT_PATH_PASSWD));
	return FAILURE;
    }

    return SUCCESS;	/* dummy */
}


/*
 * FindPTY()
 *
 * Finds the first available pseudo-terminal master/slave pair.  The master
 * side is opened and a fd returned as the function value.  A pointer to the
 * name of the slave side (i.e. "/dev/ttyp0") is returned in the argument,
 * which should be a char**.  The name itself is stored in a static buffer.
 *
 * A negative value is returned on any sort of error.
 *
 * Modified by Norstad to remove assumptions about number of pty's allocated
 * on this UNIX box.
 */
int FindPTY(p_slave_dev)
char **p_slave_dev;
{
    int master_fd;
    static char line[BUFSIZE];
    DIR *dirp;
    struct dirent *dp;

    dirp = opendir("/dev");
    while ((dp = readdir(dirp)) != NULL) {
	if (strncmp(dp->d_name, "pty", 3) == 0 && strlen(dp->d_name) == 5) {
	    strcpy(line, "/dev/");
	    strcat(line, dp->d_name);
	    if ((master_fd = open(line, O_RDWR)) >= 0) {
		line[5] = 't';
		*p_slave_dev = line;
		closedir(dirp);
		return (master_fd);
	    }
	}
    }
    closedir(dirp);
    return (-1);
}

/*
 * WriteString()
 *
 * Write a string in a single write() system call.
 */
void WriteString(fd, s)
int fd;
char *s;
{
    int l;

    l = strlen(s);
    write(fd, s, l);
}

/*
 * Talk2Child()
 *
 * Handles the conversation between the parent and child (password program)
 * processes.
 *
 * Returns SUCCESS is the conversation is completed without any problems,
 * FAILURE if any errors are encountered (in which case, it can be assumed
 * that the password wasn't changed).
 */
RESULT Talk2Child(master_fd, user, pass, newpass, emess)
int master_fd;
char *user, *pass, *newpass, *emess;
{
    char buf[BUFSIZE];
    char pswd[BUFSIZE+1];

    *emess = 0;

    TRACE(3, ("waiting for expected string (P1) from %s.\n", OPT_PATH_PASSWD));
    if (!Expect(master_fd, P1, buf)) {
	TRACE(1, ("unexpected string (P1) from %s.\n", OPT_PATH_PASSWD));
	return FAILURE;
    }
    sprintf(pswd, "%s\n", pass);
    WriteString(master_fd, pswd);

    TRACE(3, ("waiting for expected string (P2) from %s.\n", OPT_PATH_PASSWD));
    if (!Expect(master_fd, P2, buf)) {
	TRACE(1, ("unexpected string (P2) from %s.\n", OPT_PATH_PASSWD));
	return FAILURE;
    }
    sprintf(pswd, "%s\n", newpass);
    WriteString(master_fd, pswd);

    TRACE(3, ("waiting for expected string (P3) from %s.\n", OPT_PATH_PASSWD));
    if (!Expect(master_fd, P3, buf)) {
	TRACE(1, ("unexpected string (P3) from %s.\n", OPT_PATH_PASSWD));
	return FAILURE;
    }
    WriteString(master_fd, pswd);

    TRACE(3, ("waiting for expected string (P4) from %s.\n", OPT_PATH_PASSWD));
    if (!Expect(master_fd, P4, buf)) {
	TRACE(1, ("unexpected string (P4) from %s.\n", OPT_PATH_PASSWD));
	GetErrorMessage(master_fd, P4, buf);
	strcpy(emess, buf);
	return FAILURE;
    }

    while (read(master_fd, buf, sizeof(buf)) > 0);

    return SUCCESS;
}

/*
 * Match()
 *
 * Matches a string against a pattern. Wild-card characters '*' in
 * the pattern match any sequence of 0 or more characters in the string.
 * The match is case-insensitive.
 *
 * Entry: str = string.
 *        pat = pattern.
 *
 * Exit:  function result =
 *		0 if no match.
 *		1 if the string matches some initial segment of
 *		  the pattern.
 *		2 if the string matches the full pattern.
 */
int Match(str, pat)
char *str;
char *pat;
{
   int result;
   
   for (; *str && *pat && *pat != '*'; str++, pat++)
      if (tolower(*str) != tolower(*pat)) return 0;
   if (*str == 0) return *pat == 0 ? 2 : 1;
   if (*pat == 0) return 0;
   for (; *str; str++) if ((result = Match(str, pat+1)) != 0) return result;
   return 0; 
}

/*
 * Expect()
 *
 * Reads 'passwd' command output and compares it to expected output.
 *
 * Entry: master_fd = fid of master pty.
 *	  expected = pointer to array of pointers to alternate expected
 *            strings, terminated by an empty string.
 *        buf = pointer to buffer.
 *
 * Exit:  function result = SUCCESS if output matched, FAILURE if not.
 *        buf = the text read from the slave.
 *
 * Text is read from the slave and accumulated in buf. As long as
 * the text accumulated so far is an initial segment of at least 
 * one of the expected strings, the function continues the read.
 * As soon as one of full expected strings has been read, the
 * function returns SUCCESS. As soon as the text accumulated so far
 * is not an initial segment of or exact match for at least one of 
 * the expected strings, the function returns FAILURE.
 */
RESULT Expect(master_fd, expected, buf)
int master_fd;
char **expected;
char *buf;
{
    int n, m;
    char **s;
    int initialSegment;
    int result;
    char debug[BUFSIZE];
     
    n = 0;
    buf[0] = 0;
    while (1) {
	if (n >= BUFSIZE-1) {
	    TRACE(1, ("buffer overflow on read from child.\n"));
	    return FAILURE;
	}
	m = read(master_fd, buf+n, BUFSIZE-1-n);
	if (m < 0) {
	    TRACE_STRERROR(1, ("read error from child.\n"));
	    return FAILURE;
	}
	n += m;
	buf[n] = '\0';
	TRACE(3, ("'%s'\n", buf));
	initialSegment = 0;
	for (s = expected; **s != 0; s++) {
	    result = Match(buf, *s);
	    if (result == 2) {
		return SUCCESS;
	    }
	    initialSegment = initialSegment || result == 1; 
	}
	if (!initialSegment) return FAILURE;
    }
}

/*
 * GetErrorMessage()
 *
 * This function accumulates a 'passwd' command error message issued
 * after the first copy of the password has been sent.
 *
 * Entry: master_fd = fid of master pty.
 *	  expected = pointer to array of pointers to alternate expected
 *            strings for first password prompt, terminated by an 
 *            empty string.
 *        buf = pointer to buffer containing text read so far.
 *
 * Exit:  buf = the error message read from the slave.
 *
 * Text is read from the slave and accumulated in buf until the text
 * at the end of the buffer is an exact match for one of the expected
 * prompt strings. The expected prompt string is removed from the buffer,
 * returning just the error message text. Newlines in the error message
 * text are replaced by spaces.
 */
void GetErrorMessage(master_fd, expected, buf)
int master_fd;
char **expected;
char *buf;
{
    int n, m;
    char **s;
    char *p, *q;

    n = strlen(buf);
    while (1) {
	for (s = expected; **s != 0; s++) {
	    for (p = buf; *p; p++) {
		if (Match(p, *s) == 2) {
		    *p = 0;
		    for (q = buf; *q; q++) if (*q == '\n') *q = ' ';
		    return;
		}
	    }
	}
	if (n >= BUFSIZE-1) {
	    TRACE(1, ("buffer overflow on read from child.\n"));
	    return;
	}
	m = read(master_fd, buf+n, BUFSIZE+1-n);
	if (m < 0) {
	    TRACE_STRERROR(1, ("read error from child.\n"));
	    return;
	}
	n += m;
	buf[n] = 0;
    }
}

/* suitable passwd ? */
BOOL IsSuitablePasswd(passwd, user, passwd_old)
char *passwd, *user, *passwd_old;
{
    int		i;
    int		user_len, user2_len;
    int		passwd_len, passwd2_len, passwd_old_len;
    int		alpha_num, passwd_gap;
    char	user2[MAXIN], passwd2[MAXIN * 2];

    user_len		= strlen(user);
    user2_len		= user_len;
    passwd_len		= strlen(passwd);
    passwd2_len		= passwd_len * 2;
    passwd_old_len	= strlen(passwd_old);

    /* length */
    if (passwd_len < OPT_PASSLEN_MIN) {
	TRACE(1, ("new passwd is too short.\n"));
	return FALSE;
    }

    /* element */
    for (i = alpha_num = 0; i < passwd_len; i++) {
	alpha_num += (isalpha(passwd[i]) != 0);
    }
    if ((alpha_num < 2) || (alpha_num == passwd_len)) {
	TRACE(1, ("new password element is not enough.\n"));
	return FALSE;
    }

    /* user appear in passwd */
    if (user_len == passwd_len) {
	/* in order */
	strcpy(user2, user);
	strcpy(passwd2, passwd);
	strcat(passwd2, passwd);
	for (i = 0; i < user2_len; i++) {
	    user2[i] = tolower(user2[i]);
	}
	for (i = 0; i < passwd2_len; i++) {
	    passwd2[i] = tolower(passwd2[i]);
	}
	if (strstr(passwd2, user2) != NULL) {
	    TRACE(1, ("username appear in new passwd.\n"));
	    return FALSE;
	}
	/* in reverse */
	for (i = 0; i < user_len; i++) {
	    user2[i] = tolower(user[user_len - i - 1]);
	}
	user2[i] = '\0';	/* not needed, but for clearly */
	if (strstr(passwd2, user2) != NULL) {
	    TRACE(1, ("username appear in new passwd.\n"));
	    return FALSE;
	}
    }

    /* gap of old passwd */
    for (i = passwd_gap = 0; (i < passwd_len) && (i < passwd_old_len); i++) {
	passwd_gap += (tolower(passwd[i]) != tolower(passwd_old[i]));
    }
    passwd_gap += abs(passwd_len - passwd_old_len);
    if (passwd_gap < OPT_PASSDIFF) {
	TRACE(1, ("gap between old/new passwd is not enough.\n"));
	return FALSE;
    }

    return TRUE;
}


/*
 * poppassd.c
 *
 * A Eudora and NUPOP change password server.
 *
 * John Norstad
 * Academic Computing and Network Services
 * Northwestern University
 * j-norstad@nwu.edu
 *
 * Based on earlier versions by Roy Smith <roy@nyu.edu> and Daniel
 * L. Leavitt <dll.mitre.org>.
 * 
 * Doesn't actually change any passwords itself.  It simply listens for
 * incoming requests, gathers the required information (user name, old
 * password, new password) and executes /bin/passwd, talking to it over
 * a pseudo-terminal pair.  The advantage of this is that we don't need
 * to have any knowledge of either the password file format (which may
 * include dbx files that need to be rebuilt) or of any file locking
 * protocol /bin/passwd and cohorts may use (and which isn't documented).
 *
 * The current version has been tested at NU under SunOS release 4.1.2 
 * and 4.1.3, and under HP-UX 8.02 and 9.01. We have tested the server 
 * with both Eudora 1.3.1 and NUPOP 2.0.
 *
 * Other sites report that this version also works under AIX and NIS,
 * and with PC Eudora.
 *
 * Note that unencrypted passwords are transmitted over the network.  If
 * this bothers you, think hard about whether you want to implement the
 * password changing feature.  On the other hand, it's no worse than what
 * happens when you run /bin/passwd while connected via telnet or rlogin.
 * Well, maybe it is, since the use of a dedicated port makes it slightly
 * easier for a network snooper to snarf passwords off the wire.
 *
 * NOTE: In addition to the security issue outlined in the above paragraph,
 * you should be aware that this program is going to be run as root by
 * ordinary users and it mucks around with the password file.  This should
 * set alarms off in your head.  I think I've devised a pretty foolproof
 * way to ensure that security is maintained, but I'm no security expert and
 * you would be a fool to install this without first reading the code and
 * ensuring yourself that what I consider safe is good enough for you.  If
 * something goes wrong, it's your fault, not mine.
 *
 * The front-end code (which talks to the client) is directly 
 * descended from Leavitt's original version.  The back-end pseudo-tty stuff 
 * (which talks to /usr/bin/password) is directly descended from Smith's
 * version, with changes for SunOS and HP-UX by Norstad (with help from
 * sample code in "Advanced Programming in the UNIX Environment"
 * by W. Richard Stevens). The code to report /bin/passwd error messages
 * back to the client in the final 500 response, and a new version of the
 * code to find the next free pty, is by Norstad.
 *        
 * Should be owned by root, and executable only by root.  It can be started
 * with an entry in /etc/inetd.conf such as the following:
 *
 * poppassd stream tcp nowait root /usr/local/bin/poppassd poppassd
 * 
 * and in /etc/services:
 * 
 * poppassd	106/tcp
 *
 * Logs to the local2 facility. Should have an entry in /etc/syslog.conf
 * like the following:
 *
 * local2.err	/var/adm/poppassd-log
 */
 
/* Modification history.
 *
 * 06/09/93. Version 1.0.
 *
 * 06/29/93. Version 1.1.
 * Include program name 'poppassd' and version number in initial 
 *    hello message.
 * Case insensitive command keywords (user, pass, newpass, quit).
 *    Fixes problem reported by Raoul Schaffner with PC Eudora.
 * Read 'quit' command from client instead of just terminating after 
 *    password change.
 * Add new code for NIS support (contributed by Max Caines).
 *
 * 08/31/93. Version 1.2.
 * Generalized the expected string matching to solve several problems
 *    with NIS and AIX. The new "*" character in pattern strings
 *    matches any sequence of 0 or more characters.
 * Fix an error in the "getemess" function which could cause the
 *    program to hang if more than one string was defined in the
 *    P2 array.
 *
 * 03/29/94. Version 1.3.
 * Incorporated Andy Sun's changes for ULTRIX.
 *
 * 04/06/94. Version 1.4.
 * Incorporated Andy Sun's changes for OSF/1
 */

/* Andy Sun's comments about ULTRIX: (andy@ie.utoronto.ca):
 *
 * I have modified poppassd v1.2 to work under UItrix and I am keeping
 * my promise by sending you the patch (a very tiny one). The patch basically
 * does two things:
 *
 *       - added codes to Ultrix /bin/passwd prompts
 *       - use BSD4.2 syslog by specifying -DULTRIX in CFLAGS
 *
 * The real trick for getting poppassd to work under ULtrix is to add the
 * compile option "-YPOSIX". Apparently, Ultrix supports three different
 * types of termios (BSD, SYSV and POSIX) and poppassd will only work
 * by using the POSIX version of termios. So for people to compile poppassd
 * for Ultrix using the following patch, their CFLAGS and LFLAGS will have
 * to look like:
 *
 *       CFLAGS = -g -YPOSIX -DULTRIX
 *       LFLAGS = -g -YPOSIX -DULTRIX
 *
 * BTW, the machine I am using is a DECsystem 500/200 running
 * Ultrix 4.3 (Rev. 44). The patch itself should work for older versions
 * of Ultrix, but -YPOSIX cc option may vary.
 */

/* Andy Sun's comments about OSF/1:
 *
 * There is one catch in getting poppassd to work under OSF/1. I coudn't
 * get poppassd to work properly (it executes /bin/passwd alright but the
 * child process won't exit so poppassd will get stuck) under OSF/1 until
 * I ran "mkpasswd -v /etc/passwd" to create the hash database. If I don't
 * use the hash database command to create the appropriate files under /etc,
 * /bin/passwd will display the line:
 *
 *  "Hashed database not in use, only /etc/passwd text file updated."
 *
 * at the very end of /bin/passwd which is not anticpated by poppassd. I
 * tried to flush this extra output but couldn't seem to make it work.
 * So my modifications will only work on systems that have run mkpasswd.
 * This is a minor problem because all mkpasswd does is create a hash
 * table of the entries in /etc/passwd to facilitate getpwuid() and
 * getpwnam() and it quite harmless.
 *
 */

/* Steve Dorner's description of the simple protocol:
 *
 * The server's responses should be like an FTP server's responses; 
 * 1xx for in progress, 2xx for success, 3xx for more information
 * needed, 4xx for temporary failure, and 5xx for permanent failure.  
 * Putting it all together, here's a sample conversation:
 *
 *   S: 200 hello\r\n
 *   E: user yourloginname\r\n
 *   S: 300 please send your password now\r\n
 *   E: pass yourcurrentpassword\r\n
 *   S: 200 My, that was tasty\r\n
 *   E: newpass yournewpassword\r\n
 *   S: 200 Happy to oblige\r\n
 *   E: quit\r\n
 *   S: 200 Bye-bye\r\n
 *   S: <closes connection>
 *   E: <closes connection>
 */
 
